package eu.jonahbauer.raytracing.render.canvas;

import eu.jonahbauer.raytracing.render.spectral.SampledSpectrum;
import eu.jonahbauer.raytracing.render.spectral.SampledWavelengths;
import eu.jonahbauer.raytracing.render.spectral.colors.ColorRGB;
import eu.jonahbauer.raytracing.render.spectral.colors.ColorSpace;
import org.jetbrains.annotations.NotNull;

import java.awt.image.BufferedImage;
import java.util.Objects;

public final class RGBCanvas implements Canvas {
    private final int width;
    private final int height;

    private final @NotNull ColorSpace cs;
    private final @NotNull ColorRGB[][] data;

    public RGBCanvas(int width, int height, @NotNull ColorSpace cs) {
        this.width = width;
        this.height = height;
        this.cs = Objects.requireNonNull(cs);

        if (width <= 0) throw new IllegalArgumentException("width must be positive");
        if (height <= 0) throw new IllegalArgumentException("height must be positive");

        this.data = new ColorRGB[height][width];
    }

    public RGBCanvas(@NotNull BufferedImage image, @NotNull ColorSpace cs) {
        this(image.getWidth(), image.getHeight(), cs);

        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                this.data[y][x] = new ColorRGB(image.getRGB(x, y));
            }
        }
    }

    @Override
    public int getWidth() {
        return width;
    }

    @Override
    public int getHeight() {
        return height;
    }

    @Override
    public void add(int x, int y, int n, @NotNull SampledSpectrum spectrum, @NotNull SampledWavelengths lambda) {
        assert x < width;
        assert y < height;

        var rgb = spectrum.toRGB(lambda, cs);
        data[y][x] = ColorRGB.average(data[y][x], rgb, n);
    }

    @Override
    public @NotNull ColorRGB getRGB(int x, int y, @NotNull ColorSpace cs) {
        if (cs == this.cs) return get(x, y);
        return cs.toRGB(this.cs.toXYZ(get(x, y)));
    }

    public @NotNull ColorRGB get(int x, int y) {
        assert x < width;
        assert y < height;
        return Objects.requireNonNullElse(data[y][x], ColorRGB.BLACK);
    }

    public void set(int x, int y, @NotNull ColorRGB color) {
        assert x < width;
        assert y < height;
        data[y][x] = Objects.requireNonNull(color);
    }
}
