package eu.jonahbauer.raytracing.render.spectral.colors;

import eu.jonahbauer.raytracing.math.Matrix3;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.spectral.spectrum.DenselySampledSpectrum;
import eu.jonahbauer.raytracing.render.spectral.spectrum.Spectrum;
import org.jetbrains.annotations.NotNull;

import java.util.Objects;

/**
 * An RGB color space.
 */
public final class ColorSpace {
    private final @NotNull Chromaticity r;
    private final @NotNull Chromaticity g;
    private final @NotNull Chromaticity b;
    private final @NotNull Chromaticity w;
    private final @NotNull DenselySampledSpectrum illuminant;

    private final @NotNull ColorXYZ R;
    private final @NotNull ColorXYZ G;
    private final @NotNull ColorXYZ B;
    private final @NotNull ColorXYZ W;

    private final @NotNull Matrix3 XYZfromRGB;
    private final @NotNull Matrix3 RGBfromXYZ;
    private final @NotNull SpectrumTable RGBtoSpectrumTable;

    public ColorSpace(
            @NotNull Chromaticity r, @NotNull Chromaticity g, @NotNull Chromaticity b,
            @NotNull Spectrum illuminant, @NotNull SpectrumTable table
    ) {
        this.r = Objects.requireNonNull(r, "r");
        this.g = Objects.requireNonNull(g, "g");
        this.b = Objects.requireNonNull(b, "b");
        this.illuminant = new DenselySampledSpectrum(illuminant);
        this.RGBtoSpectrumTable = table;

        this.W = illuminant.toXYZ();
        this.w = W.xy();

        this.R = new ColorXYZ(r);
        this.G = new ColorXYZ(g);
        this.B = new ColorXYZ(b);
        var rgb = new Matrix3(
                R.x(), G.x(), B.x(),
                R.y(), G.y(), B.y(),
                R.z(), G.z(), B.z()
        );
        var C = rgb.invert().times(W.toVec3());

        this.XYZfromRGB = rgb.times(new Matrix3(C.x(), C.y(), C.z()));
        this.RGBfromXYZ = XYZfromRGB.invert();
    }

    public @NotNull ColorRGB toRGB(@NotNull ColorXYZ xyz) {
        var out = RGBfromXYZ.times(xyz.toVec3());
        return new ColorRGB(out.x(), out.y(), out.z());
    }

    public @NotNull ColorXYZ toXYZ(@NotNull ColorRGB rgb) {
        var out = XYZfromRGB.times(rgb.toVec3());
        return ColorXYZ.fromVec3(out);
    }

    public @NotNull Vec3 toCIELab(@NotNull ColorRGB rgb) {
        return toCIELab(toXYZ(rgb));
    }

    public @NotNull Vec3 toCIELab(@NotNull ColorXYZ xyz) {
        return new Vec3(
                116 * cieLabCbrt(xyz.y() / W.y()) - 16,
                500 * (cieLabCbrt(xyz.x() / W.x()) - cieLabCbrt(xyz.y() / W.y())),
                200 * (cieLabCbrt(xyz.y() / W.y()) - cieLabCbrt(xyz.z() / W.z()))
        );
    }

    private static double cieLabCbrt(double x) {
        var delta = 6.0 / 29.0;
        if (x > delta * delta * delta) {
            return Math.cbrt(x);
        } else {
            return x / (delta * delta * 3.0) + (4.0 / 29.0);
        }
    }

    public @NotNull SigmoidPolynomial toSpectrum(@NotNull ColorRGB rgb) {
        return RGBtoSpectrumTable.get(new ColorRGB(
                Math.max(0, rgb.r()),
                Math.max(0, rgb.g()),
                Math.max(0, rgb.b())
        ));
    }

    public @NotNull Chromaticity r() {
        return r;
    }

    public @NotNull Chromaticity g() {
        return g;
    }

    public @NotNull Chromaticity b() {
        return b;
    }

    public @NotNull Chromaticity w() {
        return w;
    }

    public @NotNull DenselySampledSpectrum illuminant() {
        return illuminant;
    }
}
