package eu.jonahbauer.raytracing.render.spectral.colors;

import eu.jonahbauer.raytracing.math.IVec3;
import eu.jonahbauer.raytracing.math.Vec3;
import org.jetbrains.annotations.NotNull;

/**
 * A CIE XYZ color
 */
public record ColorXYZ(double x, double y, double z) implements IVec3<ColorXYZ> {
    public static final double CIE_Y_INTEGRAL = 106.85689500000002;

    public static final @NotNull ColorXYZ BLACK = new ColorXYZ(0, 0, 0);
    public static final @NotNull ColorXYZ WHITE = new ColorXYZ(0, 1, 0);

    public ColorXYZ(@NotNull Chromaticity chromaticity) {
        this(chromaticity, 1);
    }

    public ColorXYZ(@NotNull Chromaticity chromaticity, double Y) {
        this(
                chromaticity.y() == 0 ? 0 : Y * chromaticity.x() / chromaticity.y(),
                chromaticity.y() == 0 ? 0 : Y,
                chromaticity.y() == 0 ? 0 : Y * (1 - chromaticity.x() - chromaticity.y()) / chromaticity.y()
        );
    }

    /*
     * Math
     */

    public static @NotNull ColorXYZ average(@NotNull ColorXYZ a, @NotNull ColorXYZ b, int index) {
        return lerp(a, b, 1d / index);
    }

    public static @NotNull ColorXYZ lerp(@NotNull ColorXYZ a, @NotNull ColorXYZ b, double t) {
        if (t < 0) return a;
        if (t > 1) return b;
        return new ColorXYZ(
                Math.fma(t, b.x - a.x, a.x),
                Math.fma(t, b.y - a.y, a.y),
                Math.fma(t, b.z - a.z, a.z)
        );
    }

    public static @NotNull ColorXYZ fma(@NotNull ColorXYZ a, @NotNull ColorXYZ b, @NotNull ColorXYZ c) {
        return new ColorXYZ(
                Math.fma(a.x, b.x, c.x),
                Math.fma(a.y, b.y, c.y),
                Math.fma(a.z, b.z, c.z)
        );
    }

    @Override
    public @NotNull ColorXYZ plus(@NotNull ColorXYZ other) {
        return new ColorXYZ(x + other.x, y + other.y, z + other.z);
    }

    @Override
    public @NotNull ColorXYZ minus(@NotNull ColorXYZ other) {
        return new ColorXYZ(x - other.x, y - other.y, z - other.z);
    }

    @Override
    public @NotNull ColorXYZ times(@NotNull ColorXYZ other) {
        return new ColorXYZ(x * other.x, y * other.y, z * other.z);
    }

    @Override
    public @NotNull ColorXYZ times(double d) {
        return new ColorXYZ(x * d, y * d, z * d);
    }

    /*
     * Vec3
     */

    @Override
    public @NotNull Vec3 toVec3() {
        return new Vec3(x, y, z);
    }

    public static @NotNull ColorXYZ fromVec3(@NotNull Vec3 vec) {
        return new ColorXYZ(vec.x(), vec.y(), vec.z());
    }

    /*
     * Accessors
     */

    public double average() {
        return (x + y + z) / 3;
    }

    public @NotNull Chromaticity xy() {
        var factor = 1 / (x + y + z);
        return new Chromaticity(factor * x, factor * y);
    }

    @Override
    public double component1() {
        return x;
    }

    @Override
    public double component2() {
        return y;
    }

    @Override
    public double component3() {
        return z;
    }
}
