package eu.jonahbauer.raytracing.render.spectral.colors;

import org.jetbrains.annotations.NotNull;

import java.io.*;
import java.util.Arrays;

/**
 * A table of sigmoid polynomials used to convert between RGB values and spectra.
 * <p>
 * The rgb values are renormalized to xyz coordinates with {@code z} being the largest of the rgb components, and
 * {@code x} and {@code y} being the other two rgb components divided by {@code z}. By this construction, {@code x},
 * {@code y} and {@code z} are all in the range [0, 1] which allows for better use of the samples in a fixed grid.
 * The {@code z} coordinate is additionally remapped using {@link #zNodes} to improve sampling at both ends of the scale.
 * <p>
 * The coefficients of the sigmoid functions are stored in a flattened five-dimensional array with indices as described
 * in {@link #coefficients}.
 */
public final class SpectrumTable {
    private final int resolution;

    /**
     * the remapped {@code z} values
     */
    private final double[] zNodes;

    /**
     * the stored coefficients as a flattened five-dimensional array with the following indices
     * <ol>
     *     <li>the component index of the biggest rgb component</li>
     *     <li>the {@code z} coordinate</li>
     *     <li>the {@code y} coordinate</li>
     *     <li>the {@code x} coordinate</li>
     *     <li>the coefficient index</li>
     * </ol>
     */
    private final double[] coefficients;

    public static void write(@NotNull SpectrumTable table, @NotNull OutputStream out) throws IOException {
        var dos = new DataOutputStream(out);
        dos.writeInt(table.resolution);
        for (double z : table.zNodes) {
            dos.writeDouble(z);
        }
        for (double c : table.coefficients) {
            dos.writeDouble(c);
        }
        dos.flush();
    }

    public static @NotNull SpectrumTable read(@NotNull InputStream in) throws IOException {
        var dis = new DataInputStream(in);
        var resolution = dis.readInt();
        var nodes = new double[resolution];
        for (int i = 0; i < resolution; i++) {
            nodes[i] = dis.readDouble();
        }
        var table = new double[3 * resolution * resolution * resolution * 3];
        for (int i = 0; i < table.length; i++) {
            table[i] = dis.readDouble();
        }
        return new SpectrumTable(resolution, nodes, table);
    }

    SpectrumTable(int resolution, double @NotNull[] zNodes, double[] coefficients) {
        this.resolution = resolution;
        this.zNodes = zNodes;
        this.coefficients = coefficients;

        // check input array lengths
        if (zNodes.length != resolution) {
            throw new IllegalArgumentException("length of zNodes must be equal to the RESOLUTION");
        }
        if (coefficients.length != 3 * resolution * resolution * resolution * 3) {
            throw new IllegalArgumentException("coefficients length must be 3 * RESOLUTION * RESOLUTION * RESOLUTION * 3");
        }
        // check ascending zNodes
        for (int i = 1; i < resolution; i++) {
            if (zNodes[i - 1] >= zNodes[i]) {
                throw new IllegalArgumentException("zNodes must be in increasing order");
            }
        }
        if (zNodes[0] != 0.0 || zNodes[zNodes.length - 1] != 1.0) {
            throw new IllegalArgumentException("zNodes must start with 0.0 and end with 1.0");
        }
    }

    public @NotNull SigmoidPolynomial get(@NotNull ColorRGB color) {
        // handle uniform rgb values
        if (color.r() == color.g() && color.g() == color.b()) {
            return new SigmoidPolynomial(0, 0, (color.r() - .5) / Math.sqrt(color.r() * (1 - color.r())));
        }

        // find maximum component and compute remapped component values
        var max = color.r() > color.g()
                ? (color.r() > color.b() ? 0 : 2)
                : (color.g() > color.b() ? 1 : 2);

        var z = color.get(max);
        var x = color.get((max + 1) % 3) * (resolution - 1) / z;
        var y = color.get((max + 2) % 3) * (resolution - 1) / z;

        // compute integer indices and offsets for coefficient interpolation
        int xi = Math.min((int) x, resolution - 2);
        int yi = Math.min((int) y, resolution - 2);
        int zi = Arrays.binarySearch(zNodes, z);
        if (zi < 0) {
            zi = -zi - 2;
        } else if (zi > 0) {
            zi = zi - 1;
        }

        var dx = x - xi;
        var dy = y -yi;
        var dz = (z - zNodes[zi]) / (zNodes[zi + 1] - zNodes[zi]);

        // trilinearly interpolate sigmoid polynomial coefficients
        var c = new double[3];
        for (int i = 0; i < 3; i++) {
            c[i] = lerp(dz,
                        lerp(dy,
                             lerp(dx, get(max, zi + 0, yi + 0, xi + 0, i), get(max, zi + 0, yi + 0, xi + 1, i)),
                             lerp(dx, get(max, zi + 0, yi + 1, xi + 0, i), get(max, zi + 0, yi + 1, xi + 1, i))
                        ),
                        lerp(dy,
                             lerp(dx, get(max, zi + 1, yi + 0, xi + 0, i), get(max, zi + 1, yi + 0, xi + 1, i)),
                             lerp(dx, get(max, zi + 1, yi + 1, xi + 0, i), get(max, zi + 1, yi + 1, xi + 1, i))
                        )
            );
        }

        return new SigmoidPolynomial(c[0], c[1], c[2]);
    }

    private double get(int l, int z, int y, int x, int i) {
        return coefficients[(((l * resolution + z) * resolution + y) * resolution + x) * 3 + i];
    }

    private static double lerp(double t, double a, double b) {
        return Math.fma(t, b, Math.fma(-t, a, a));
    }
}
