package eu.jonahbauer.raytracing.render.spectral.colors;

import eu.jonahbauer.raytracing.math.Matrix3;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.spectral.spectrum.Spectrum;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Objects;
import java.util.stream.IntStream;

/**
 * Generates a lookup table for RGB to spectrum conversion.
 * <p>
 * The spectrum for each RGB value is a {@link SigmoidPolynomial} with coefficients such that the round trip error
 * from converting the RGB value to a spectrum and back is minimized.
 * <p>
 * <img src="doc-files/rgb2spectrum.png">
 */
public final class SpectrumTableGenerator {
    private static final double EPSILON = 1e-4;
    private static final int ITERATIONS = 15;

    private final int resolution = 64;
    private final @NotNull ColorSpace cs;

    public static void main(String[] args) throws IOException {
        var generator = new SpectrumTableGenerator(ColorSpaces.DCI_P3);
        var table = generator.generate();

        try (var out = Files.newOutputStream(Path.of("DCI_P3_spectrum.bin"))) {
            SpectrumTable.write(table, out);
        }
    }

    public SpectrumTableGenerator(@NotNull ColorSpace cs) {
        this.cs = Objects.requireNonNull(cs);
    }

    public @NotNull SpectrumTable generate() {
        var scale = new double[resolution];
        for (int i = 0; i < scale.length; i++) {
            var t = (double) i / (resolution - 1);
            scale[i] = smoothstep(smoothstep(t));
        }

        var table = new double[3 * resolution * resolution * resolution * 3];

        for (int l0 = 0; l0 < 3; l0++) {
            var l = l0;
            IntStream.range(0, resolution).parallel().forEach(i -> {
                System.out.println("l = " + l + ", i = " + i);
                var x = (double) i / (resolution - 1);
                for (int j = 0; j < resolution; j++) {
                    var y = (double) j / (resolution - 1);

                    var start = resolution / 5;

                    var c = new double[3];
                    for (int k = start; k < resolution; k++) {
                        var z = scale[k];
                        var idx = ((((l * resolution + k) * resolution) + j) * resolution + i) * 3;
                        var color = getColor(l, x, y, z);
                        generate(color, c, table, idx);
                    }

                    Arrays.fill(c, 0);
                    for (int k = start; k >= 0; --k) {
                        var z = scale[k];
                        var idx = ((((l * resolution + k) * resolution) + j) * resolution + i) * 3;
                        var color = getColor(l, x, y, z);
                        generate(color, c, table, idx);
                    }
                }
            });
        }
        return new SpectrumTable(resolution, scale, table);
    }

    private void generate(@NotNull ColorRGB rgb, double @NotNull[] c, double @NotNull[] out, int offset) {
        gaussNewton(rgb, c, ITERATIONS);
        double c0 = 360.0, c1 = 1.0 / (830.0 - 360.0);
        double A = c[0], B = c[1], C = c[2];
        out[offset] = A * c1 * c1;
        out[offset + 1] = B * c1 - 2 * A * c0 * c1 * c1;
        out[offset + 2] = C - B * c0 * c1 + A * c0 * c0 * c1 * c1;
    }

    /**
     * Use Gauss-Newton algorithm to calculate coefficients {@code c} of a {@link SigmoidPolynomial} such that the round
     * trip error from converting the {@code rgb} value to a spectrum and back is minimized.
     * @param rgb the input color
     * @param c the coefficients, used as initial values and output
     * @param it the number of iterations
     */
    private void gaussNewton(@NotNull ColorRGB rgb, double @NotNull[] c, int it) {
        var bestQuality = Double.POSITIVE_INFINITY;
        var bestCoefficients = new double[3];

        for (int i = 0; i < it; ++i) {
            var polynomial = new SigmoidPolynomial(c[0], c[1], c[2]);
            var residual = getResidual(rgb, polynomial);
            var jacobian = getJacobian(rgb, polynomial);

            var delta = jacobian.decompose(1e-15).solve(residual);
            for (int j = 0; j < 3; ++j) {
                c[j] -= delta.get(j);
            }

            // catch runaway
            double max = Math.max(Math.max(c[0], c[1]), c[2]);
            if (max > 200) {
                for (int j = 0; j < 3; ++j) {
                    c[j] *= 200 / max;
                }
            }

            var quality = residual.squared();
            if (quality <= 1e-6) {
                return;
            } else if (quality < bestQuality) {
                bestQuality = quality;
                System.arraycopy(c, 0, bestCoefficients, 0, 3);
            }
        }

        System.arraycopy(bestCoefficients, 0, c, 0, 3);
    }

    /**
     * Calculates the Jacobian matrix of the {@code polynomial}.
     */
    private @NotNull Matrix3 getJacobian(@NotNull ColorRGB rgb, @NotNull SigmoidPolynomial polynomial) {
        var jac = new double[3][3];

        // central finite difference coefficients for first derivative with sixth-order accuracy
        var factors = new double[] { -1d/60, 3d/20, -3d/4, 0, 3d/4, -3d/20,	1d/60 };

        for (int i = 0; i < 3; i++) {
            var derivative = Vec3.ZERO;
            for (int d = - factors.length / 2, j = 0; j < factors.length; d++, j++) {
                if (factors[j] == 0) continue;
                var tmp = switch (i) {
                    case 0 -> new SigmoidPolynomial(polynomial.c0() + d * EPSILON, polynomial.c1(), polynomial.c2());
                    case 1 -> new SigmoidPolynomial(polynomial.c0(), polynomial.c1() + d * EPSILON, polynomial.c2());
                    case 2 -> new SigmoidPolynomial(polynomial.c0(), polynomial.c1(), polynomial.c2() + d * EPSILON);
                    default -> throw new AssertionError();
                };
                var r = getResidual(rgb, tmp);
                derivative = Vec3.fma(factors[j], r, derivative);
            }

            for (int j = 0; j < 3; j++) {
                jac[j][i] = derivative.get(j) / EPSILON;
            }
        }

        return new Matrix3(
                jac[0][0], jac[0][1], jac[0][2],
                jac[1][0], jac[1][1], jac[1][2],
                jac[2][0], jac[2][1], jac[2][2]
        );
    }

    /**
     * Calculates the difference between the RGB color and the result of converting the RGB color to a spectrum using
     * the given coefficients, illuminating it with the color space's standard illuminant, and converting it back to an
     * RBG color. The output is a vector in CIE Lab color space.
     */
    private @NotNull Vec3 getResidual(@NotNull ColorRGB rgb, @NotNull SigmoidPolynomial polynomial) {
        var out = new SigmoidPolynomialSpectrum(polynomial, cs).toXYZ();
        return cs.toCIELab(rgb).minus(cs.toCIELab(out));
    }


    private static double smoothstep(double x) {
        return x * x * (3.0 - 2.0 * x);
    }

    private static @NotNull ColorRGB getColor(int l, double x, double y, double z) {
        var rgb = new double[3];
        rgb[l] = z;
        rgb[(l + 1) % 3] = x * z;
        rgb[(l + 2) % 3] = y * z;
        return new ColorRGB(rgb[0], rgb[1], rgb[2]);
    }

    private record SigmoidPolynomialSpectrum(@NotNull SigmoidPolynomial polynomial, @NotNull ColorSpace cs) implements Spectrum {

        @Override
        public double max() {
            return polynomial.max();
        }

        @Override
        public double get(double lambda) {
            var l = (lambda - Spectrum.LAMBDA_MIN) / (Spectrum.LAMBDA_MAX - Spectrum.LAMBDA_MIN);
            return polynomial.get(l) * cs.illuminant().get(lambda);
        }
    }
}
