package eu.jonahbauer.raytracing.render.texture;

import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.spectral.colors.ColorRGB;
import eu.jonahbauer.raytracing.render.spectral.colors.ColorSpace;
import eu.jonahbauer.raytracing.render.spectral.spectrum.RGBAlbedoSpectrum;
import eu.jonahbauer.raytracing.render.spectral.spectrum.RGBIlluminantSpectrum;
import eu.jonahbauer.raytracing.render.spectral.spectrum.Spectrum;
import org.jetbrains.annotations.NotNull;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.Objects;

public final class ImageTexture implements Texture {
    private static final String PATH_PREFIX = "/eu/jonahbauer/raytracing/textures/";

    private final int width;
    private final int height;
    private final @NotNull Spectrum[][] spectra;

    public ImageTexture(@NotNull BufferedImage image, @NotNull ColorSpace cs, @NotNull Type type, boolean gamma) {
        this.width = image.getWidth();
        this.height = image.getHeight();
        this.spectra = new Spectrum[height][width];

        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                var rgb = new ColorRGB(image.getRGB(x, y));
                if (gamma) rgb = ColorRGB.inverseGamma(rgb);
                spectra[y][x] = type.newSpectrum(cs, rgb);
            }
        }
    }

    public ImageTexture(@NotNull String path, @NotNull ColorSpace cs) {
        this(read(path), cs, Type.ALBEDO, true);
    }

    private static @NotNull BufferedImage read(@NotNull String path) {
        try (var in = Objects.requireNonNull(ImageTexture.class.getResourceAsStream(PATH_PREFIX + path))) {
            return ImageIO.read(in);
        } catch (IOException ex) {
            throw new UncheckedIOException(ex);
        }
    }

    @Override
    public @NotNull Spectrum get(double u, double v, @NotNull Vec3 p) {
        u = Math.clamp(u, 0, 1);
        v = 1 - Math.clamp(v, 0, 1);
        int x = (int) (u * (width - 1));
        int y = (int) (v * (height - 1));
        return spectra[y][x];
    }

    public enum Type {
        ALBEDO {
            @Override
            protected @NotNull Spectrum newSpectrum(@NotNull ColorSpace cs, @NotNull ColorRGB rgb) {
                return new RGBAlbedoSpectrum(cs, rgb);
            }
        },
        ILLUMINANT {
            @Override
            protected @NotNull Spectrum newSpectrum(@NotNull ColorSpace cs, @NotNull ColorRGB rgb) {
                return new RGBIlluminantSpectrum(cs, rgb);
            }
        },
        ;

        protected abstract @NotNull Spectrum newSpectrum(@NotNull ColorSpace cs, @NotNull ColorRGB rgb);
    }
}
